/*
 * bg_jobs - LKM 
 * --------------------
 * Background Jobs
 */

#include <linux/module.h>  
#include <linux/kernel.h>  
#include <linux/fs.h>
#include <linux/file.h>
#include <linux/namei.h>
#include <linux/fs.h>
#include <linux/proc_fs.h>
#include <linux/uaccess.h>
#include <linux/mutex.h>
#include <linux/list.h>
#include <linux/completion.h>
#include <linux/sched.h>
#include <linux/version.h>

/* Include the right file according to kernel version  */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,27)
#include <asm/semaphore.h>
#else
#include <linux/semaphore.h>
#endif

MODULE_AUTHOR("Donato Capitella");
MODULE_DESCRIPTION ("bg_jobs");
MODULE_LICENSE("GPL");

/* Name of the stop job used to signal the kernel thread to exit */
#define STOP_JOB "STOP"

/*
 * Job structure
 */
#define JOB_NAME_MAX_SIZE 25
struct job {
	struct list_head list;
	char *name;					/* name of the job */
	void (*do_job)(void *args);	/* function that executes the job */
	void *args;					/* pointer to arguments to be passed to do_job() */
};

/*
 * List of jobs
 */
LIST_HEAD(job_queue);

/*
 * Synchronisation structures for producer/consumer problem
 */
DEFINE_MUTEX(mutex);
struct semaphore available = __SEMAPHORE_INITIALIZER(available, 0); 

DECLARE_COMPLETION(thread_exit);

/**
 * build_job - allocates and return a new job with given arguments.
 */
struct job *build_job(char *job_name, void (*do_job)(void *args), void *args) 
{
	struct job *new = kmalloc(sizeof(struct job), GFP_KERNEL);
	if (new) {
		new->name = job_name;
		new->do_job = do_job;
		new->args = args;
	}
	return new;
}

/**
 * dequeue - extracts a job from the queue and returns it.
 */
struct job *dequeue(void) {
	struct job *head = NULL;

	down(&available);	
	mutex_lock(&mutex);
	head = list_entry(job_queue.next, struct job, list);
	list_del(&head->list);
	mutex_unlock(&mutex);	
	
	return head;
}

/**
 * enqueue - inserts a job in the queue.
 */
int enqueue(struct job *job) {
	int ret = -1;
	if(job) {
		mutex_lock(&mutex);
		list_add_tail(&(job->list), &job_queue);
		mutex_unlock(&mutex);
		up(&available);
		ret = 0;
	}
	return ret;
}

/**
 * kthr_worker - body of the worker thread that executes the jobs in background.
 */
static int kthr_worker(void *unused)
{
	struct job *job;
	
	/* needed to detach the kernel thread from user resources */
	daemonize("kthr_worker");
	
	/*
	 * Main loop of the worker thread.
	 */
	for(;;) {
		/* wait for next job */
		job = dequeue();
		
		/* Check if we need to stop */
		if (!strcmp(job->name, STOP_JOB)) {
			break;
		}
		
		/* Execute Job */
		printk(KERN_INFO "Executing Job with name: %s\n", job->name);	
		job->do_job(job->args);
		kfree(job);
	}
	
	printk(KERN_INFO "[kthr_worker] Exiting...\n");	
	complete_and_exit(&thread_exit, 0);
}

/*
 * sleep_job: sleeps for 
 */
void sleep_job(void *secs)
{
	int seconds = (int)secs;
	/* Pretend to do some job */
	printk(KERN_INFO "Going to sleep for %d seconds...\n", seconds);
	set_current_state(TASK_INTERRUPTIBLE);
	schedule_timeout (HZ * seconds);
	printk(KERN_INFO "I am awake now!\n");
}

void hello_job(void *unused)
{
	printk(KERN_INFO "Hello Job!!! ;-)\n");
}

/*
 * init_module
 */
int init_module(void)
{
	/* pupulate the job list */
	printk(KERN_INFO "Generating jobs...\n");	
	enqueue(build_job("JOB #1", hello_job, NULL));
	enqueue(build_job("JOB #2", sleep_job, (void*)5));
	enqueue(build_job("JOB #3", sleep_job, (void*)2));
	enqueue(build_job("JOB #4", sleep_job, (void*)3));
	enqueue(build_job("JOB #5", hello_job, NULL));

	/* start a kernel thread that will execute the jobs in background */
	printk(KERN_INFO "Starting kernel thread...\n");	
	if (!kernel_thread(kthr_worker, NULL, CLONE_KERNEL)) {
		printk(KERN_INFO "Failed to generate a kernel thread\n");
		return -1;
	}
	
	return 0;
}


void cleanup_module(void)
{
	/* send STOP job and wait for thread completion */
	enqueue(build_job(STOP_JOB, NULL, NULL));
	wait_for_completion(&thread_exit);
}  


