/*
 * counter_nosync 
 * --------------------
 * Counter module without synchronization
 */

#include <linux/module.h>  /* Needed by all modules */
#include <linux/kernel.h>  /* Needed for KERN_ALERT */
#include <linux/fs.h>
#include <linux/file.h>
#include <linux/namei.h>
#include <linux/fs.h>
#include <linux/proc_fs.h>
#include <linux/uaccess.h>
#include <linux/spinlock.h>

MODULE_AUTHOR("Donato Capitella");
MODULE_DESCRIPTION ("counter_spinlock");
MODULE_LICENSE("GPL");

/* 
 * Name of the /proc entry the module uses to communicate 
 * with user processes
 */
#define PROC_ENTRY_NAME "counter_spinlock"

/*
 * /proc entry descriptor
 */
static struct proc_dir_entry *counter_entry;

/*
 * Counter variable, initialized to 0
 */
static long counter = 0;

/*
 * LOCK to protect counter
 */
spinlock_t lock = SPIN_LOCK_UNLOCKED;

/*
 * counter_read
 *
 * used to transfer the results of the
 * computations to the user process.
 */
static int 
counter_read(char *buf, char **start, off_t offset, 
				int len, int *eof, void *unused)
{	
	long tmp;
	*eof = 1;
	
	/* Read the value of counter into temporary variable */
	spin_lock(&lock);
	tmp = counter;
	spin_unlock(&lock);
	
	return sprintf(buf, "%li", tmp);
}

/*
 * counter_write
 *
 * Read next digit to add to counter
 */
static int 
counter_write(struct file *file, const char *buffer, 
				unsigned long count, void *data)
{
	char digit_char;
	int digit;

	/* Each command can be only a single ASCII digit */
	if (count != 1) {
		return -EFAULT;
	}

	/* copy data from user space */
	if(copy_from_user(&digit_char, buffer, count)) {
		return -EFAULT;
	}
	 
	/* Check if input is a valid digit */
	if (digit_char < '0' || digit_char > '9') {
		return -EFAULT;		
	}
	
	/* convert to integer */
	digit = digit_char - '0';

	/* Add the number */
	spin_lock(&lock);
	counter += digit;
	spin_unlock(&lock);

	return count;
}

/*
 * init_module - initialize /proc entry
 */
int init_module(void)
{
	counter_entry = create_proc_entry(PROC_ENTRY_NAME, S_IRUGO | S_IWUGO, NULL);
	counter_entry->read_proc = counter_read;
	counter_entry->write_proc = counter_write;
	return 0;
}


void cleanup_module(void)
{
	if (counter_entry) {
		remove_proc_entry(PROC_ENTRY_NAME, NULL);
	}
}  


